import invariant from 'invariant';
import addEvent from '@thinkeloquent/rd2020-dom/src/element/addEvent';
import addClass from '@thinkeloquent/rd2020-dom/src/element/addClass';
import removeClass from '@thinkeloquent/rd2020-dom/src/element/removeClass';
import addAttribute from '@thinkeloquent/rd2020-dom/src/element/addAttribute';
import getTargetFromEvent from '@thinkeloquent/rd2020-dom/src/element/getTargetFromEvent';
import { findOneWithDocument } from '@thinkeloquent/rd2020-dom/src/selector/findOne';
import { findWithDocument } from '@thinkeloquent/rd2020-dom/src/selector/find';
import getTagName from '@thinkeloquent/rd2020-dom/src/element/getTagName';

import getNodeTypeFromElement, { DOCUMENT_NODE } from '@thinkeloquent/rd2020-dom/src/element/getNodeTypeFromElement';
import stylesheet from './index.less';

export const IGNORE_ELEMENT_ATTR = 'hightlight-ignore-element';

export const IGNORE_ELEMENT_REACT_ATTR = { 'hightlight-ignore-element': true };

export const HIGHLIGHT_IGNORE_ATTR = 'hightlight-ignore-element';

let DEBUG = false;
export const setDebug = (boolean) => DEBUG = Boolean(boolean);

export const highlightClassName = `root-${process.env.BUILD_ID}`;
export const highlightSelectedClassName = `selected-${process.env.BUILD_ID}`;
export const headStyleTagsId = `style-${process.env.BUILD_ID}`;

export const fromDocument = (document) => ({
  body: document.body,
  head: document.head,
  document,
});

export const fromIframe = (document) => ({
  body: document.body,
  head: document.head,
  document,
});

export const logger = (...args) => {
  if(!DEBUG) return;
  console.log('ELEMENT_HIGHLIGHT', ...args);
};

class ElementHighlight {
  constructor(element) {
    const isDocument = getNodeTypeFromElement(element).nodeType === DOCUMENT_NODE;
    const {
      body,
      head,
      document,
    } = isDocument ? fromDocument(element) : fromIframe(element);
    
    this.dom = {
      body,
      head,
      document,
    }

    this.previousHoveredElement = null;
    this.previousSelectedElement = null;
    this.disposeOf = [];

    this.classes = {
      ignoreElement: stylesheet.locals.highlight,
      highlight: stylesheet.locals.highlight,
      highlightSelected: stylesheet.locals.highlightSelected,
    };

    this.insertStyleSheet();
  }

  setIgnoreAttribute(element) {
    if(element) {
      element.setAttribute(HIGHLIGHT_IGNORE_ATTR, "true")
    }
  }

  useStyleSheet() {
    if(stylesheet.use) stylesheet.use();
  }

  unuseStyleSheet() {
    if(stylesheet.unuse) stylesheet.unuse();
  }

  insertStyleSheet() {
    this.stylesheet = document.createElement('style');
    this.stylesheet.id = headStyleTagsId;
    this.stylesheet.innerHTML = stylesheet.toString();
    this.dom.head.insertAdjacentHTML('beforeend', this.stylesheet.outerHTML);
  }

  removeClassName(className) {
    let currentlySelected = findWithDocument(this.dom.document, `.${className}`);

    if (currentlySelected && currentlySelected.length) {
        removeClass(currentlySelected, className);
    }
    return currentlySelected;
  }

  getSelected() {
    return findOneWithDocument(this.dom.document, this.classes.highlightSelected);
  }

  ignoreElement(query) {
    const elements = findWithDocument(this.dom.document, query);

    if(!elements) {
      logger(`ignoreElement element not found: ${query}`);
      return;
    }
    addAttribute(elements, IGNORE_ELEMENT_ATTR, "true")
  }

  stopPropagation(query) {
    const element = findOneWithDocument(this.dom.document, query);

    if(!element) {
      logger(`stopPropagation element not found: ${query}`);
      return;
    }

    addEvent(element, 'click', (event) => {
      logger(`ignore element ${query}`);
      event.preventDefault();
      event.stopPropagation();
    });
  }


  disposal() {
    for(let index in this.disposeOf) {
      this.disposeOf[index]();
      delete this.disposeOf[index];
    }
  }

  cleanup() {
    this.removeHighlight()
    this.disposal();
  }

  removeHighlight() {
    this.previousHoveredElement = null;
    this.previousSelectedElement = null;
    this.removeClassName(this.classes.highlight);
    this.removeClassName(this.classes.highlightSelected);
  }


  // const mouseMoveHandler = (e) => {
  //   /**
  //    * Since Chrome may trigger redundant mousemove event evne if
  //    * we didn't really move the mouse, we should make sure that
  //    * mouse coordinates really changed then respond to mousemove
  //    * event
  //    * @see https://code.google.com/p/chromium/issues/detail?id=327114
  //    */
  //   if (((e.pageX || e.clientX) === this._prevX && (e.pageY || e.clientY) === this._prevY) ||
  //       (this._prevX === null && this._prevY === null)) {
  //     return false;
  //   }
  
  //   this.handleMouseMove.call(this, e);
  // };

  mouseover({
    editableTags,
    onHoverIgnore,
  }) {
    this.disposeOf.push(
      addEvent(this.dom.body, 'mouseover', (event) => {
        const target = getTargetFromEvent(event);
        const tagName = getTagName(target);
        const ignore = target.getAttribute(IGNORE_ELEMENT_ATTR);
        
        if(ignore === "true") {
          logger(`IGNORE_ELEMENT_ATTR element mouseover: ${tagName}`);
          if(typeof onHoverIgnore === 'function') onHoverIgnore(target, this.previousSelectedElement);
          return;
        }

        if(editableTags.indexOf(tagName) === -1) {
          logger(`ignore element mouseover: ${tagName}`);
          if(typeof onHoverIgnore === 'function') onHoverIgnore(target, this.previousSelectedElement);
          return;
        }

        addClass(target, this.classes.highlight);
        this.previousHoveredElement = target;
      })
    )
  }

  mouseout() {
    this.disposeOf.push(
      addEvent(this.dom.body, 'mouseout', (event) => {
        const element = event ? event.target : this.previousHoveredElement;
        if (element) removeClass(element, this.classes.highlight);
      })
    )
  }

  click(options) {
    const {
      onClickIgnore,
      editableTags,
      onReSelect,
    } = options;
    this.disposeOf.push(
      addEvent(this.dom.body, 'click', (event) => {
        const target = getTargetFromEvent(event);
        const tagName = getTagName(target);
        const ignore = target.getAttribute(IGNORE_ELEMENT_ATTR);

        if(ignore === "true") {
          logger(`IGNORE_ELEMENT_ATTR element click: ${tagName}`);
          if(typeof onClickIgnore === 'function') onClickIgnore(target, this.previousSelectedElement);
          this.returnValue = true;
          return true;
        }
        
        event.preventDefault();
        event.stopPropagation();

        if(this.previousSelectedElement === target) {
          if(typeof onReSelect === 'function') onReSelect(this.previousSelectedElement);
          logger(`reselect ignore click: ${tagName}`);
          return;
        }

        if(editableTags.indexOf(tagName) === -1) {
          logger(`ignore element click: ${tagName}`);
          if(typeof onClickIgnore === 'function') onClickIgnore(target, this.previousSelectedElement);
          return;
        }
        
        this.selectElement(target, options);
      })
    )
  }

  selectElement(target, {
    onSelect,
    onRemove,
  }) {
    if(this.previousSelectedElement) {
      if(typeof onRemove === 'function') onRemove(target, this.previousSelectedElement);
      removeClass(this.previousSelectedElement, this.classes.highlightSelected);
    }

    addClass(target, this.classes.highlightSelected);

    if(typeof onSelect === 'function') onSelect(target, this.previousSelectedElement);

    this.previousSelectedElement = target;
  }

  initialize(argOptions = {}) {
    const options = Object.assign({}, argOptions);
    invariant(Array.isArray(options.editableTags), 'editableTags array is missing');
    this.mouseover(options);
    this.mouseout(options);
    this.click(options);
  }

  getData() {
    return {}
  }

}

export default ElementHighlight;